package PureMessage::Database::Schema;

use strict;
use Carp qw(carp croak);
use HTTP::Date qw(str2time time2iso);
use PureMessage::Database::Config;
use DBI;
use ActiveState::Run qw/run/;
use PureMessage::Version qw/$PREFIX/;

# Identifies database schema version. Schema version should be updated
# (incremented) if the schema undergoes changes but there should be no
# more than one version update since the last release. In other words,
# if you're making changes to this file, then do update the version, but
# only if it hasn't already been updated since the last release that has
# been made available to the customers (this includes Beta releases).
# Schema changes include structural changes (table/column changes),
# index changes, changes in stored procedures and triggers, and changes
# in constant data.
#
# Refer to the schema_changes_howto.txt in the top-level directory of
# the PureMessage::Database PPM package for more details about schema
# version changes, schema testing and database conversion scripts.
our $VERSION = '6.405';
my $PGBIN = "$PREFIX/postgres/bin";

# This module defines PMX database schema.
#
# Graphical representation of main tables and their relationships
# (quarantine portion only):
#
#  ---------------         ---------------         ---------------
# |               |      /|               |\      |               |
# | message       |-------| quarantine_   |-------| quarantine_   |
# |               |      \| rule_hit      |/      | rule          |
#  --------------- \       ---------------         ---------------
#         |         \
#         |          \
#         |           \
#        /|\           \
#  ---------------      \
# |               |      \
# | message_      |       \
# | recipient     |        \
#  ---------------          \
#        \|/                 \
#         |                   \
#         |                    \-----------------------------------
#         |                    |                                   |
#  ---------------             |                                   |
# |               |            |           message_body            |
# | recipient     |            |                                   |
# |               |            |                                   |
#  ---------------              -----------------------------------


# A future revision will involve rebuilding this module to handle separating the
# DB into N components, and splitting the schema to match.   
sub new {
    my $class = shift;
    my $dsn = shift;
    my %opts;
    if (ref($dsn) eq 'HASH') {
        %opts = %$dsn;
        $dsn = $opts{dsn};

    }

    my $self = bless {}, $class;

    if ($dsn) {
	croak("Unrecognized DSN '$dsn'") unless $dsn =~ /^dbi:(\w+)/;
        ($self->{dbname} = $dsn) =~ s#.+dbname=([^;]+);.+#$1#;
    }
    else {
        $self->{store} = $opts{store} || "pmx_quarantine";
        my $pmdb_cfg;
        eval {
            $pmdb_cfg = PureMessage::Database::Config->new();
            $dsn = $pmdb_cfg->store_value($self->{store}, "dsn");
        };
        unless ($dsn) {
            # only for testing
            $dsn = $ENV{PMX_DBI_SOURCE};
            croak("Undefined DSN") unless $dsn;
        }

        ($self->{dbname} = $dsn) =~ s#.+dbname=([^;]+);.+#$1#;
    }

    $self->{dsn} = $dsn;
    $self->{sql_dir} = "$PREFIX/etc/data/sql";

    return $self;
}

# connects to the specified database and creates db objects
sub init {
    my $self = shift;
    my $store_id = shift;

    # Only initialze the schema if we can't find an existing schema version
    # OR: If we find a schema version, this DB is already initialized
    my $ver;
    eval { $ver = $self->current_version };
    unless ($ver) {
        run("$PGBIN/psql -qtf $PREFIX/etc/data/sql/schema.sql $self->{dbname}");
    }

    my $pmdb_cf = PureMessage::Database::Config->new(writable => 1);
    $pmdb_cf->store_value($store_id, 'dsn', $self->{dsn})
        unless $pmdb_cf->store_value($store_id, 'dsn');
    $pmdb_cf->save;
}

# Drop the pmx_quarantine database
sub drop {
    my $self = shift;
    run("-$PGBIN/dropdb pmx_quarantine");
    run("$PGBIN/createdb -E SQL_ASCII pmx_quarantine");
}

sub current_version {
    my $self = shift;
    my $schema_sql = <<"__SQL__";
select description
from   pg_description where objoid =
           (select oid
            from pg_namespace
            where nspname = 'public');
__SQL__
    my $version = qx($PGBIN/psql -qtc "$schema_sql" $self->{dbname});
    if ($version =~ m/PMX (\d\.\d{3})/) {
        return $1;
    }
    else {
        die "unable to retrieve current database version - can not proceed";
    }

    return $1;
}

# Finds and runs all of the sql update scripts
sub convert {
    my $self = shift;
    my $target_version = shift || $VERSION;
    my @conversion_script = $self->get_conversion_scripts($target_version);
    foreach my $script (@conversion_script) {
        run("$PGBIN/psql -qtf $script $self->{dbname}");
    }
}

# returns an ordered list of SQL database conversion scripts that
# correspond to this particular conversion (from current version to
# target version)
sub get_conversion_scripts {
    my $self = shift;
    my $target_version = shift;
    my $current_version = $self->current_version;
    my $sql_dir = $self->{sql_dir};
    opendir(DIR, $sql_dir) || die "unable to open $sql_dir: $!";
    my @script = sort grep /[\d.]+_to_[\d.]+m?\.sql/, readdir(DIR);
    closedir(DIR);
    my @script_needed;

    # check if they concatenate all the way
    my ($version_before, $version_after) = ($current_version, "");;
    foreach my $script (@script) {
        my $from_to = $script;
        $from_to =~ s/m?\.sql//;
        my ($from, $to) = split '_to_', $from_to;

        # if it succeeds the most recent one than we need it
        # also, if it brings us to the desired version
        # we don't need to look any more
        if ($from eq $version_before) {
            $version_after = $to;
            push @script_needed, "$sql_dir/$script";
            last if ($to eq $target_version);
            $version_before = $to;
        }
    }
    unless ($version_after eq $target_version) {
        die "can not find (all) SQL scripts required for conversion " .
            "from $current_version to $target_version - unable to proceed";
    }

    return @script_needed;
}

1;
